@@ -261,12 +261,12 @@ static const struct {
 	{ L"\\wave", L"ヾ(＾∇＾)" },
 };
 
-void inputCompleteAdd(void) {
+void inputCompletion(void) {
 	char mbs[256];
 	for (size_t i = 0; i < ARRAY_LEN(Macros); ++i) {
 		size_t n = wcstombs(mbs, Macros[i].name, sizeof(mbs));
 		assert(n != (size_t)-1);
-		completeAdd(None, mbs, Default);
+		completePush(None, mbs, Default);
 	}
 }
 
@@ -296,15 +296,16 @@ static struct {
 	size_t pos;
 	size_t len;
 	bool suffix;
+	struct Cursor curs;
 } tab;
 
 static void tabAccept(void) {
-	completeAccept();
+	completeAccept(&tab.curs);
 	tab.len = 0;
 }
 
 static void tabReject(void) {
-	completeReject();
+	completeReject(&tab.curs);
 	tab.len = 0;
 }
 
@@ -332,9 +333,9 @@ static int tabComplete(struct Edit *e, uint id) {
 		tab.suffix = true;
 	}
 
-	const char *comp = complete(id, tab.pre);
+	const char *comp = completePrefix(&tab.curs, id, tab.pre);
 	if (!comp) {
-		comp = complete(id, tab.pre);
+		comp = completePrefix(&tab.curs, id, tab.pre);
 		tab.suffix ^= true;
 	}
 	if (!comp) {
@@ -467,7 +468,7 @@ static void keyCtrl(wchar_t ch) {
 		break; case L'I': error = tabComplete(edit, windowID());
 		break; case L'J': inputEnter();
 		break; case L'K': error = editFn(edit, EditDeleteTail);
-		break; case L'L': clearok(curscr, true);
+		break; case L'L': clearok(curscr, true); wrefresh(curscr);
 		break; case L'N': windowShow(windowNum() + 1);
 		break; case L'P': windowShow(windowNum() - 1);
 		break; case L'R': windowSearch(editString(edit, &buf, &cap, NULL), -1);
